#include <stdio.h>
#include <poll.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <syslog.h>
#include <sys/signal.h>
#include <unistd.h>

#define GPIO_FN_MAXLEN	32
#define POLL_TIMEOUT	10000
#define RDBUF_LEN	5

int nodaemon=0;
static char *usage = 
"Usage: shutkey -g <GPIO> [-D] [-v]\n"
"options:\n"
" -g <number>   specify a gpio number (default: 8)\n"
"               This is NOT the physical pin number of the connector!\n"
"               (the gpio 8 is physical pin number 24)\n"
" -v            verbose / debug mode\n"
" -D            don't fork, run as a foreground process\n"
"\n";

void hup_handler(int sig) {
  if (nodaemon==0) {
    syslog(LOG_USER|LOG_INFO, "shutkey closed");
  }

  exit(EXIT_SUCCESS);
}

logging(int loglevel, char* logstring) {
  if (nodaemon==1) {
    printf("%s\n",logstring);
  }
  if (nodaemon==0) {
    syslog(LOG_USER|loglevel, "%s", logstring);
  }
}
                
int main(int argc, char **argv) {
  char fn[GPIO_FN_MAXLEN];
  int fd,ret;
  struct pollfd pfd;
  char rdbuf[RDBUF_LEN];
  int i = 0;
  int gpionumber=8;
  static int debug = 0;
  FILE *fp;
  char errstr[128];

  memset(rdbuf, 0x00, RDBUF_LEN);
  memset(fn, 0x00, GPIO_FN_MAXLEN);

  for (i = 1; i < argc; i++){
    if (!(argv[i][0] == '-' || argv[i][0] == '/') || (strlen(argv[i]) <= 1)){
      fprintf(stdout, "Unknown option : %s\n", argv[i]);
      /* Show usage */
      fputs(usage, stdout);
      return -1;
    }
    int moreargs = (i < argc - 1);
    switch(argv[i][1]){
      case 'g':
        gpionumber = moreargs ? atoi(argv[++i]) : gpionumber;
        break;
      case 'D':
        nodaemon=1;
        break;
      case 'v':
        debug = 1;
        break;
      default:
        fprintf(stdout, "Unknown option : %s\n", argv[i]);
        /* Show usage */
        fputs(usage, stdout);
        return -1;
    }
  }
  if (debug > 0){
    fprintf(stdout, "Running with parameters:\n");
    fprintf(stdout, "\tdebug=%d\n", debug);
    fprintf(stdout, "\tgpionumber=%d\n", gpionumber);
    fprintf(stdout, "\tnodaemon=%d\n", nodaemon);
  }

  openlog("shutkey",LOG_PID, LOG_USER);

  if (nodaemon==0) {
    /* Our process ID and Session ID */
    pid_t pid, sid;
    /* Fork off the parent process */
    pid = fork();
    if (pid < 0) {
      logging(LOG_CRIT,"unable to fork - exiting");
      exit(EXIT_FAILURE);
    }
    /* If we got a good PID, then we can exit the parent process. */
    if (pid > 0) { // Child can continue to run even after the parent has finished executing
      logging(LOG_INFO,"fork ok");
      exit(EXIT_SUCCESS);
    }
    /* Change the file mode mask */
    umask(0);
    /* start logging */
    syslog(LOG_USER|LOG_INFO, "shutkey started");
    /* Create a new SID for the child process */
    sid = setsid();
    if (sid < 0) {
      logging(LOG_CRIT,"unable to create a child SID - exiting");
      exit(EXIT_FAILURE);
    }
    /* Change the current working directory */
    if ((chdir("/")) < 0) {
      logging(LOG_CRIT,"unable to change the working directory - exiting");
      exit(EXIT_FAILURE);
    }
    /* Close out the standard file descriptors */
    //Because daemons generally dont interact directly with user so there is no need of keeping these open
    close(STDIN_FILENO);
    close(STDOUT_FILENO);
    close(STDERR_FILENO);
  }
  
  signal(SIGHUP, hup_handler);
  signal(SIGINT, hup_handler);
  signal(SIGALRM, hup_handler);
  signal(SIGTERM, hup_handler);

  fp=fopen("/sys/class/gpio/export", "w");
  if( fp != NULL ) {
    fprintf(fp,"%d",gpionumber);
    fclose(fp);
  } else {
    logging(LOG_CRIT,"unable to write to /sys/class/gpio/export - exiting");
    exit(EXIT_FAILURE);
  }
  
  snprintf(fn, GPIO_FN_MAXLEN-1, "/sys/class/gpio/gpio%d/edge", gpionumber);
  fp=fopen(fn, "w");
  if( fp != NULL ) {
    fprintf(fp,"falling");
    fclose(fp);
  } else {
    snprintf(errstr, 63, "unable to write to /sys/class/gpio/gpio%d/edge - exiting", gpionumber);
    logging(LOG_CRIT,errstr);
    exit(EXIT_FAILURE);
  }

  snprintf(fn, GPIO_FN_MAXLEN-1, "/sys/class/gpio/gpio%d/value", gpionumber);
  fd=open(fn, O_RDONLY);
  if(fd<0) {
    snprintf(errstr, 63, "%s - exiting", fn);
    logging(LOG_CRIT,errstr);
    exit(EXIT_FAILURE);
  }
  pfd.fd=fd;
  pfd.events=POLLPRI;

  ret=read(fd, rdbuf, RDBUF_LEN-1);
  if(ret<0) {
    logging(LOG_CRIT,"error reading gpio - exiting");
    exit(EXIT_FAILURE);
  }

  while(1) {
    memset(rdbuf, 0x00, RDBUF_LEN);
    lseek(fd, 0, SEEK_SET);
    ret=poll(&pfd, 1, POLL_TIMEOUT);
    if(ret<0) {
      logging(LOG_CRIT,"error polling interrupt - exiting");
      close(fd);
      exit(EXIT_FAILURE);
    }
    if(ret==0) {
      continue;
    }
    ret=read(fd, rdbuf, RDBUF_LEN-1);
    if(ret<0) {
      logging(LOG_CRIT,"error reading gpio - exiting");
      exit(EXIT_FAILURE);
    }
    logging(LOG_INFO,"got interrupt from shutdown key, shutting down now");
    system("shutdown -P now");
    exit(0);
  }
  close(fd);
  return 0;
}

